-- CPSC 312 - Project 2
-- by Khurram Ali Jaffery

-- Main Components:
-- minimax algorithm
-- a board evaluator
-- state search
-- movement generators (and by extension, tree generator, new state generator)
-- crusher
-- custom data types (already done)

-- Piece is a data representation of possible pieces on a board
-- where D is an empty spot on the board
--       W is a piece of the White player
--       B is a piece of the Black player
--

data Piece = D | W | B deriving (Eq, Show)

-- An enum for the different directions that a piece can move.
data Direction = TL | TR | L | R | BL | BR deriving (Enum, Bounded)

--
-- Point is a tuple of 2 elements
-- representing a point on a grid system
-- where the first element represents the x coordinate
--       the second element represents the y coordinate
--

type Point = (Int, Int)

--
-- Tile is a tuple of 2 elements 
-- representing what a point is occupied by
-- where the first element represents a piece 
--       the second element represents a point
--

type Tile  = (Piece, Point)

--
-- Board is a list of Pieces, thus it is an internal representation
-- of the provided string representation of the board, it maintains
-- the same order as the string representation of the board
--

type Board = [Piece]

--
-- Grid is a list of Points, thus it is an internal representation
-- of the hexagonal grid system translated into a coordinate 
-- system to easily maintain and make moves on the board
--

type Grid = [Point]

--
-- State is a list of Tile, thus it is an internal representation precisely
-- for the purposes of zipping the board and the grid together in order
-- to keep easier track of the effects on the pieces of making moves on grid
--

type State = [Tile]

--
-- Next is a data representation for storing and passing around information within
-- the tree generating function, allowing it to correctly generate new children
-- 
-- Next consists of 4 elements
-- where usedDepth is an integer reprsenting the current depth level
--       newBoard is the next board to add to the tree
--       seenBoards is the updated history to avoid possible future trouble boards
--       cplayer is the current player for whom the board was generated for
--

data Next a = Next {usedDepth :: Int, newBoard :: a, seenBoards :: [a], cplayer :: Piece}

--
-- Tree is a data representation for the search tree, it is an extention of 
-- the rose tree widely used for implementing such unequally branched search trees
--
-- Tree consists of 3 elements
-- where depth is an integer representing the depth level of the node
--       board is the game state at that node
--       nextBoards are the child nodes of the current node
--

data Tree a = Node {depth :: Int, board :: a, nextBoards :: [Tree a]} deriving (Show)

--
-- BoardTree is the internal representation of the search tree of the given board
-- that is to be generatated for correctly implementing the minimax algorithm.
--

type BoardTree = Tree Board

--
-- Slide is a tuple of 2 elements
-- an internal representation of a slide
-- where the first element represents the point to move from
--       the second element represents the adjacent point to move to
--

type Slide = (Point,Point)

--
-- Jump is a tuple of 2 elements
-- an internal representation of a leap
-- where the first element represents the point to move from
--       the second element represents the adjacent point to move over
--       the third element represents the point to move to
--

type Jump = (Point,Point,Point)

--
-- Move is a tuple of 2 elements
-- an internal representation of a move
-- where the first element represents the point to move from
--       the second element represents the point to move to
--
-- Note: in essence it is the same as a slide however the idea
--       is that a jump can be reduced to a move as in effect 
--       nothing happens the point moved over in a jump
--

type Move = (Point,Point)

--
-- Some test results to see what functions are producing 
--
-- run = crusher ["W------------BB-BBB","----W--------BB-BBB","-W-----------BB-BBB"] 'W' 2 3
grid0 = generateGrid 3 2 4 []
slides0 = generateSlides grid0 3
jumps0 = generateLeaps grid0 3
board0 = sTrToBoard "WWW-WW-------BB-BBB"
newBoards0 = generateNewStates board0 [] grid0 slides0 jumps0 W
tree0 = generateTree board0 [] grid0 slides0 jumps0 W 4 3
board1 = sTrToBoard "WW--W----B--B-B----"
newBoards1 = generateNewStates board1 [] grid0 slides0 jumps0 W
tree1 = generateTree board1 [] grid0 slides0 jumps0 W 4 3

grid2 = generateGrid 2 1 2 []
slides2 = generateSlides grid2 2
jumps2 = generateLeaps grid2 2
board2 = sTrToBoard "WW---BB"
newBoards2 = generateNewStates board2 [] grid2 slides2 jumps2 W
tree2 = generateTree board2 [] grid2 slides2 jumps2 W 10 2
-- heuristic0 = boardEvaluator W [] 3

--
-- crusher
--
-- This function consumes a list of boards, a player, the depth of 
-- search tree, the size of the provide boards, and produces the 
-- next best board possible for the provided player, and accordingly
-- makes the move and returns new board consed onto the list of boards
--
-- Arguments:
-- -- (current:old): current represents the most recent board, old is
--                   the history of all boards already seen in game
-- -- p: 'W' or 'B' representing the player the program is
-- -- d: an Integer indicating depth of search tree
-- -- n: an Integer representing the dimensions of the board
--
-- Returns: a list of String with the new current board consed onto the front
--

crusher :: [String] -> Char -> Int -> Int -> [String]
crusher (current:old) p d n = boardToStr ((stateSearch (sTrToBoard current) (map sTrToBoard old)
    grid
    (generateSlides grid n)
    (generateLeaps grid n)
    ((\ p -> if p == 'W' then W else B)p) d n)):(current:old)
    where grid = (generateGrid n (n-1) (2*(n-1)) [])

--
-- gameOver
--
-- This function consumes a board, a list of boards, and the dimension
-- of board and determines whether the given board is in a state where
-- the game has ended by checking if the board is present in the provided
-- list of boards or either the W or B pieces are less than dimension of board
--
-- Arguments:
-- -- board: a Board representing the most recent board
-- -- history: a list of Boards of representing all boards already seen
-- -- n: an Integer representing the dimensions of the board
--
-- Returns: True if the board is in a state where the game has ended, otherwise False
--

gameOver :: Board -> [Board] -> Int -> Bool
gameOver board history n =
    elem board history || occurencesOf B board < n || occurencesOf W board < n
    where occurencesOf e lst = foldl (\ l r -> if r == e then l + 1 else l) 0 lst

--
-- sTrToBoard
--
-- This function consumes a list of characters which can be either 'W' or 'B'
-- or '-' and converts them to a list of pieces, i.e W or B or D respectively
--
-- Arguments:
-- -- s: the String to convert into piece-wise representation
--
-- Note: This function would convert "WWW-WW-------BB-BBB" to
--       [W,W,W,D,W,W,D,D,D,D,D,D,D,B,B,D,B,B,B]
--
-- Returns: the Board corresponding to the string
--

sTrToBoard :: String  -> Board
sTrToBoard s = map (\ x -> check x) s
    where 
        check 'W' = W
        check 'B' = B
        check '-' = D

--
-- boardToStr
--
-- This function consumes a board which is a list of either W or B  or D and 
-- converts them to a list of characters, i.e 'W' or 'B' or 'D' respectively
--
-- Arguments:
-- -- b: the Board to convert into char-wise representation
--
-- Note: This function would convert [W,W,W,D,W,W,D,D,D,D,D,D,D,B,B,D,B,B,B] 
--       to "WWW-WW-------BB-BBB"
--
-- Returns: the String corresponding to the board 
--

boardToStr :: Board -> String
boardToStr b = map (\ x -> check x) b
    where 
        check W = 'W'
        check B = 'B'
        check D = '-'

--
-- generateGrid
--
-- This function consumes three integers (described below) specifying how to
-- properly generate the grid and also a list as an accumulator; to generate a
-- regular hexagon of side length n, pass n (n- 1) (2 * (n - 1)) and []
--
-- Arguments:
-- -- n1: one more than max x-coordinate in the row, initialized always to n
-- -- n2: the number of rows away from the middle row of the grid
-- -- n3: the current y-coordinate i.e the current row number
-- -- acc: an accumulator that keeps track of accumulating rows of grid 
--         initialized to []
--
-- Note: This function on being passed 3 2 4 [] would produce
--       [(0,0),(1,0),(2,0)
--        (0,1),(1,1),(2,1),(3,1)
--        (0,2),(1,2),(2,2),(3,2),(4,2)
--        (0,3),(1,3),(2,3),(3,3)
--        (0,4),(1,4),(2,4)]
--
-- Returns: the corresponding Grid i.e the acc when n3 == -1
--

generateGrid :: Int -> Int -> Int -> Grid -> Grid
generateGrid n1 n2 n3 acc 
    | n3 == -1      = acc
    | otherwise     = generateGrid nn1 (n2 - 1) (n3 - 1) (row ++ acc)
        where
            row = map (\ x -> (x,n3)) [0 .. (n1 - 1)]
            nn1 = if n2 > 0 then n1 + 1 else n1 - 1

-- Finds the adjacent point for the given point in the given direction.
adjacentPointInDir :: Point -> Int -> Direction -> Point
adjacentPointInDir (x,y) n dir = 
    case dir of
        TL -> if y < n then (x-1, y-1) else (x, y-1)
        TR -> if y < n then (x, y-1) else (x+1, y-1)
        L -> (x-1, y)
        R -> (x+1, y)
        BL -> if y >= (n - 1) then (x-1, y+1) else (x, y+1)
        BR -> if y >= (n - 1) then (x, y+1) else (x+1, y+1)

-- Finds adjacent points for the given point.
--
-- Arguments:
-- -- p: The point for which we are returning the adjacent points.
-- -- n: An Integer representing the dimensions of the grid.
--
-- Returns: A list of all adjacent points to the given point.
adjacentPoints :: Point -> Int -> [Point]
adjacentPoints p n = map (\ dir -> adjacentPointInDir p n dir) [(minBound :: Direction) ..]

-- Finds valid adjacent points for the given point.
--
-- Arguments:
-- -- p: The point for which we are returning the adjacent points.
-- -- grid: The grid.
-- -- n: An Integer representing the dimensions of the grid.
--
-- Returns: A list of all valid adjacent points to the given point.
validAdjacentPoints :: Point -> Grid -> Int -> [Point]
validAdjacentPoints p grid n = filter (\ e -> elem e grid) (adjacentPoints p n)

--
-- generateSlides
--
-- This function consumes a grid and the size of the grid, accordingly
-- generates a list of all possible slides from any point on the grid to
-- any adjacent point on the grid
--
-- Arguments:
-- -- b: the Grid to generate slides for 
-- -- n: an Integer representing the dimensions of the grid
-- 
-- Note: This function is only called at the initial setup of the game, 
--       it is a part of the internal representation of the game, this 
--       list of all possible slides is only generated once; and when 
--       generating next moves, the program decides which slides out of 
--       all these possible slides could a player actually make
--
-- Returns: the list of all Slides possible on the given grid
--

generateSlides :: Grid -> Int -> [Slide]
generateSlides b n = foldl (\ slides pt ->
    (map (\ e -> (pt,e)) (validAdjacentPoints pt b n)) ++ slides) [] b

-- Finds the leap for the given point in the given direction.
getLeapInDir :: Point -> Int -> Direction -> Jump
getLeapInDir p n dir = (p, apInDir, adjacentPointInDir apInDir n dir)
        where apInDir = adjacentPointInDir p n dir

-- Finds leaps for the given point.
--
-- Arguments:
-- -- p: The point for which we are returning the leaps.
-- -- n: An Integer representing the dimensions of the grid.
--
-- Returns: A list of tuples representing leaps from the given point.
findLeaps :: Point -> Int -> [Jump]
findLeaps p n = map (\ dir -> getLeapInDir p n dir) [(minBound :: Direction) ..]

-- Finds all valid leaps for the given point.
--
-- Arguments:
-- -- p: The point for which we are returning the leaps.
-- -- grid: The grid.
-- -- n: An Integer representing the dimensions of the grid.
--
-- Returns: A list of all valid leaps from the given point.
findValidLeaps :: Point -> Grid -> Int -> [Jump]
findValidLeaps p grid n = (filter
                            (\ (_,e1,e2) -> (elem e2 grid) && (elem e1 grid)) -- Not sure if we even need to check e1
                            (findLeaps p n))

--
-- generateLeaps
--
-- This function consumes a grid and the size of the grid, accordingly
-- generates a list of all possible leaps from any point on the grid over
-- any adjacent point on the grid to any point next to the adjacent point
-- such that it is movement in the same direction
--
-- Arguments:
-- -- b: the Grid to generate leaps for 
-- -- n: an Integer representing the dimensions of the grid
-- 
-- Note: This function is only called at the initial setup of the game, 
--       it is a part of the internal representation of the game, this 
--       list of all possible leaps is only generated once; and when 
--       generating next moves, the program decides which leaps out of 
--       all these possible leaps could a player actually make
--
-- Returns: the list of all Jumps possible on the given grid
--

generateLeaps :: Grid -> Int -> [Jump]
generateLeaps b n = foldl (\ leaps pt -> (findValidLeaps pt b n) ++ leaps) [] b

--
-- stateSearch
--
-- This function consumes the arguments described below, based on the internal
-- representation of the game, if there is no point in playing the game as the
-- current board is in a state where the game has ended then just return the 
-- board, else generate a search tree till the specified depth and apply 
-- minimax to it by using the appropriately generated heuristic
--
-- Arguments:
-- -- board: a Board representing the most recent board
-- -- history: a list of Boards of representing all boards already seen
-- -- grid: the Grid representing the coordinate-grid the game being played
-- -- slides: the list of all Slides possible for the given grid
-- -- jumps: the list of all Jumps possible for the given grid
-- -- player: W or B representing the player the program is
-- -- depth: an Integer indicating depth of search tree
-- -- num: an Integer representing the dimensions of the board
--
-- Returns: the current board if game is over, 
--          otherwise produces the next best board
--

stateSearch :: Board -> [Board] -> Grid -> [Slide] -> [Jump] -> Piece -> Int -> Int -> Board
stateSearch board history grid slides jumps player depth num =
    if (gameOver board history num)
        then board
        else minimax (generateTree board history grid slides jumps player depth num) (boardEvaluator player history num)

--
-- generateTree
--
-- This function consumes the arguments described below, and builds a search
-- tree till specified depth from scratch by using the current board and
-- generating all the next states recursively; however it doesn't generate
-- children of those states which are in a state where the game has ended.
--
-- Arguments:
-- -- board: a Board representing the most recent board
-- -- history: a list of Boards of representing all boards already seen
-- -- grid: the Grid representing the coordinate-grid the game being played
-- -- slides: the list of all Slides possible for the given grid
-- -- jumps: the list of all Jumps possible for the given grid
-- -- player: W or B representing the player the program is
-- -- depth: an Integer indicating depth of search tree
-- -- n: an Integer representing the dimensions of the board
--
-- Returns: the corresponding BoardTree generated till specified depth
--
-- Node {depth :: Int, board :: a, nextBoards :: [Tree a]}

generateTree :: Board -> [Board] -> Grid -> [Slide] -> [Jump] -> Piece -> Int -> Int -> BoardTree
generateTree board _ _ _ _ _ 0 _ = Node 0 board []
generateTree board history grid slides jumps player depth n = Node depth board
    ((\ b h dim -> 
    if gameOver b h dim then []
    else (map (\ newBoard -> generateTree newBoard (board:history) grid slides jumps ((\ p -> if p == W then B else W) player) (depth - 1) n)
    (generateNewStates board history grid slides jumps player))
    ) board history n)

--
-- generateNewStates
--
-- This function consumes the arguments described below, it first generates a
-- list of valid moves, applies those moves to the current board to generate 
-- a list of next boards, and then checks whether or not that move would 
-- have been possible by filtering out those boards already seen before
--
-- Arguments:
-- -- board: a Board representing the most recent board
-- -- history: a list of Boards of representing all boards already seen
-- -- grid: the Grid representing the coordinate-grid the game being played
-- -- slides: the list of all Slides possible for the given grid
-- -- jumps: the list of all Jumps possible for the given grid
-- -- player: W or B representing the player the program is
--
-- Returns: the list of next boards
--
generateNewStates :: Board -> [Board] -> Grid -> [Slide] -> [Jump] -> Piece -> [Board]
generateNewStates board history grid slides jumps player = foldl (\ z move ->
    if (elem (applyMove board grid move player) history) then z
    else (applyMove board grid move player):z)
    [] (moveGenerator (boardToState' board grid) slides jumps player)

--
-- applyMove
--
-- Takes a board and its grid, a move, and the current player and outputs the next board after the move is executed
--
applyMove :: Board -> Grid -> Move -> Piece -> Board
applyMove [] _ _ _ = []
applyMove (boardH:boardR) (gridH:gridR) (moveStart, moveEnd) player =
    if gridH == moveStart then D:applyMove boardR gridR (moveStart, moveEnd) player
    else if gridH == moveEnd then player:applyMove boardR gridR (moveStart, moveEnd) player
    else boardH:applyMove boardR gridR (moveStart, moveEnd) player
    
--
-- boardToState
--
-- This function consumes a board and its dimension and converts into a state
--
boardToState :: Board -> Int -> State
boardToState b n = boardToState' b (generateGrid n (n-1) (2*(n-1)) [])

boardToState' :: Board -> Grid -> State
boardToState' b g = zip b g

--
-- checkPointInState
--
-- This function consumes a state and a point and returns the piece at that point
--
checkPointInState :: State -> Point -> Piece
checkPointInState [] _ = D
checkPointInState ((tpc, tpt):sr) p = if tpt == p then tpc else checkPointInState sr p

--
-- moveGenerator
--
-- This function consumes a state, a list of possible jumps, 
-- a list of possible slides and a player from whose perspective 
-- to generate moves, to check which of these jumps and slides 
-- the player could actually make, and produces a list of valid moves
--
-- Arguments:
-- -- state: a State representing the most recent state
-- -- slides: the list of all Slides possible for the given grid
-- -- jumps: the list of all Jumps possible for the given grid
-- -- player: W or B representing the player the program is
--
-- Note: This is the only instance where the program makes use of the
--       type State, for our purposes it is zipping the board and the
--       grid together for making it easier to make moves.
--
-- Note:
-- -- oP is opponentsPieces
-- -- pP is playersPieces
-- -- vS is validSlides
-- -- vJ is validJumps
--
-- Returns: the list of all valid moves that the player could make
--
moveGenerator :: State -> [Slide] -> [Jump] -> Piece -> [Move]
moveGenerator state slides jumps player = concat [moveGeneratorSlides state state slides player, moveGeneratorJumps state state jumps player]

-- Helper for moveGenerator that handles moves from slides
moveGeneratorSlides :: State -> State -> [Slide] -> Piece -> [Move]
moveGeneratorSlides [] _ _ _ = []
moveGeneratorSlides ((tilePiece, tilePoint):str) ref sl pi = concat((foldl
    (\ z (slideStart, slideEnd) -> if tilePoint == slideStart && tilePiece == pi && ((checkPointInState ref slideEnd) == D) then (slideStart, slideEnd):z else z) [] sl):[moveGeneratorSlides str ref sl pi])

-- Helper for moveGenerator that handles moves from jumps
moveGeneratorJumps :: State -> State -> [Jump] -> Piece -> [Move]
moveGeneratorJumps [] _ _ _ = []
moveGeneratorJumps ((tilePiece, tilePoint):str) ref sl pi = concat((foldl
    (\ z (jumpStart, jumpOver, jumpEnd) -> if tilePoint == jumpStart && tilePiece == pi && ((checkPointInState ref jumpOver) == pi) && ((checkPointInState ref jumpEnd) /= pi) then (jumpStart, jumpEnd):z else z) [] sl):[moveGeneratorJumps str ref sl pi])
--
-- boardEvaluator
--
-- This function consumes a board and performs a static board evaluation, by 
-- taking into account whose perspective the program is playing from, the list 
-- of boards already seen, the size of the board, and whether or not it is the
-- program's turn or not; to generate quantitative measures of the board, and 
-- accordingly produce a goodness value of the given board 
--
-- Arguments:
-- -- player: W or B representing the player the program is
-- -- history: a list of Boards of representing all boards already seen
-- -- n: an Integer representing the dimensions of the board
-- -- board: a Board representing the most recent board
-- -- myTurn: a Boolean indicating whether it is the program's turn or the opponents.
--
-- Returns: the goodness value of the provided board
--

boardEvaluator :: Piece -> [Board] -> Int -> Board -> Bool -> Int
boardEvaluator player history n board myTurn = 
    if elem board history then
        if myTurn then (-2*n) -- -2n for losing
        else (2*n) -- 2n for winning
    else if (occurencesOf player board) < n then (-2*n) -- -2n for losing
    else if (occurencesOf ((\ p -> if p == W then B else W) player) board) < n then (2*n) -- 2n for winning
    else (occurencesOf player board) - (occurencesOf ((\ p -> if p == W then B else W) player) board)
    where occurencesOf e lst = foldl (\ l r -> if r == e then l + 1 else l) 0 lst
--
-- minimax
--
-- This function implements the minimax algorithm, it consumes a search tree, 
-- and an appropriate heuristic to apply to the tree, by applying minimax it
-- produces the next best board that the program should make a move to
--
-- Arguments:
-- -- (Node _ b children): a BoardTree to apply minimax algorithm on
-- -- heuristic: a paritally evaluated boardEvaluator representing the
--               appropriate heuristic to apply based on the size of the board,
--               who the program is playing as, and all the boards already seen
--
-- Returns: the next best board
--

minimax :: BoardTree -> (Board -> Bool -> Int) -> Board
minimax (Node _ b ((Node ha hb hc):childrenR)) heuristic = snd(foldl (\ (zS, zB) (Node ca cb cc) ->
    if minimax' (Node ca cb cc) heuristic True > zS then (minimax' (Node ca cb cc) heuristic True, cb)
    else (zS, zB)) (minimax' (Node ha hb hc) heuristic True, hb) childrenR)
--
-- minimax'
--
-- This function is a helper to the actual minimax function, it consumes 
-- a search tree, an appropriate heuristic to apply to the leaf nodes of 
-- the tree, and based on whether it would have been the maximizing 
-- player's turn, it accordingly propogates the values upwards until
-- it reaches the top to the base node, and produces that value.
--
-- Arguments:
-- -- (Node _ b []): a BoardTree
-- -- (Node _ b children): a BoardTree
-- -- heuristic: a paritally evaluated boardEvaluator representing the
--               appropriate heuristic to apply based on the size of the board,
--               who the program is playing as, and all the boards already seen
-- -- maxPlayer: a Boolean indicating whether the function should be maximizing
--               or miniziming the goodness values of its children
--
-- Returns: the minimax value at the top of the tree
--

minimax' :: BoardTree -> (Board -> Bool -> Int) -> Bool -> Int
minimax' (Node _ b []) heuristic maxPlayer = heuristic b maxPlayer;
minimax' (Node _ b (childrenH:childrenR)) heuristic True = foldl (\ z child -> max (minimax' child heuristic False) z) (minimax' childrenH heuristic False) childrenR
minimax' (Node _ b (childrenH:childrenR)) heuristic False = foldl (\ z child -> min (minimax' child heuristic True) z) (minimax' childrenH heuristic True) childrenR

